from sendgrid import SendGridAPIClient
from .exceptions import MailSendFailure, MalformedConfiguration, BadSenderObject
from pattan_sendgrid.models import Config

class PattanEmail:
    """
    Useful SendGrid API abstraction for sending emails.
    It includes a command line interface to construct the configuration file by querying the SendGrid API directly.
    """
    def __init__(self, config_json=None ):
        if not config_json:
            raise MalformedConfiguration
        try:
            # pydantic validator makes sure each property is defined and has a default value set
            self.config = Config.model_validate_json(config_json)
        except Exception as e:
            raise MalformedConfiguration

        self.api_key = self.config.api_key
        self.ip_pool = self.config.ip_pools
        self.unsubscribe_groups = self.config.unsubscribe_groups
        self.senders = self.config.senders
        self.templates = self.config.email_templates
        self.sg = SendGridAPIClient(api_key=self.api_key)



    def send_template_email(self, to_addr, dynamic_template_data=None, sender='DEFAULT', 
                            email_template="DEFAULT", asm_group="DEFAULT", ip_pool="DEFAULT"):
        """
        Send the same email to one or more recipients.
        :param to_addr: email address dict or list of address dicts e.g. [{'name':'bob', 'email':'bob@example.com'}],
        :param dynamic_template_data: dict that defines all the variables used in the selected email_template
        :param email_template: string Name of the template you want to use e.g. "PaTTAN Standard Template"
        :param sender: string Name of the sender email address, e.g. "no-reply@PaTTAN"; or email address dict
        :param asm_group: string Name of the asm group (a.k.a. unsubscribe group)  e.g. "SendGrid Tech Test Group"
        :param ip_pool: string Name of the ip_pool e.g. "Pattan_Transactional"
        :return: SendGrid client response or throws an exception
        """

        from_email = self.get_from_email_from_sender(sender)

        # the to_addr can be a list of just a string or an email address.
        if isinstance(to_addr, str):
            to_addr = [{'name': to_addr, 'email': to_addr}]

        personalizations = [{
            'to': to_addr,
            'dynamic_template_data': dynamic_template_data,
        }]

        asm = {
            'group_id': self.unsubscribe_groups[asm_group].id,
            'groups_to_display': [
                self.unsubscribe_groups[asm_group].id
            ]
        }

        template_id = self.templates[email_template].id

        message = {
            "from": from_email,
            "personalizations": personalizations,
            "template_id": template_id,
            "asm": asm,
            "ip_pool_name": self.ip_pool[ip_pool].name,
        }

        try:
            sg_response = self.sg.client.mail.send.post(request_body=message)
        except Exception as e:
            raise MailSendFailure
        return sg_response

    def send_personalized_template_email(self, personalization_list, template_id, sender='DEFAULT',
                                         asm_group="DEFAULT", ip_pool="DEFAULT"):
        """
        This function should be used when the email is customized for each recipient.
        :param personalization_list: contains a sender tuple and all the parameters in the sendgrid template.
        :return:
        """
        from_email = self.get_from_email_from_sender(sender)

        asm = {
            'group_id': self.unsubscribe_groups[asm_group].id,
            'groups_to_display': [
                self.unsubscribe_groups[asm_group].id
            ]
        }

        message = {
            'asm': asm,
            'from': from_email,
            'ip_pool_name': self.ip_pool[ip_pool].name,
            'template_id': template_id,
            'personalizations': personalization_list
        }

        try:
            sg_response = self.sg.client.mail.send.post(request_body=message)
        except Exception as e:
            raise MailSendFailure
        return sg_response


    def get_from_email_from_sender(self, sender):
        if isinstance(sender, str):
            if sender not in self.senders.keys():
                raise MalformedConfiguration('Assigned sender is not defined in your configuration')

            sender = self.senders[sender]
            from_email = {'email': sender.from_address.email}
            if sender.nickname:
                from_email['name'] = sender.nickname
            
            return from_email
        
        else:
            if 'email' in sender:
                from_email = {'email': sender['email']}
                if 'name' in sender:
                    from_email['name'] = sender['name']
                
                return from_email
            
            else:
                raise BadSenderObject



from pydantic import BaseModel, model_validator
from typing import Dict, List


class From(BaseModel):
    email: str
    name: str


class Sender(BaseModel):
    from_address: From
    nickname: str
    reply_to: From
    address: str
    address_2: str
    city: str
    state: str
    zip: str


class UnSubscribeGroup(BaseModel):
    id: int


class EmailTemplate(BaseModel):
    id: str
    name: str
    variables: List[str]

class IpPool(BaseModel):
    name: str

class Config(BaseModel):
    api_key: str
    senders: Dict[str, Sender]
    ip_pools: Dict[str, IpPool]
    unsubscribe_groups: Dict[str, UnSubscribeGroup]
    email_templates: Dict[str, EmailTemplate]

    @model_validator(mode='before')
    def check_default_key(cls, values):
        if 'DEFAULT' not in values['senders']:
            raise ValueError("Senders missing a default")
        if 'DEFAULT' not in values['ip_pools']:
            raise ValueError("ip_pools missing a default")
        if 'DEFAULT' not in values['unsubscribe_groups']:
            raise ValueError("unsubscribe_groups missing a default")
        if 'DEFAULT' not in values['email_templates']:
            raise ValueError("email_templates missing a default")
        return values


import click
import json
import re

@click.command()
@click.option('--default-sender', help='Sender label as defined in sendgrid. If left unset the first one found will be set as the default ')
@click.option('--default-ip-pool', help='Sendgird -> settings -> ip addresses . If left unset the first one found will be set as the default ')
@click.option('--default-unsubscribe_group', help='Sendgrid -> marketing -> unsubscribe group . If left unset the first one found will be set as the default ')
@click.option('--default-dynamic-template', help='Sendgrid -> email api -> dynamic template . If left unset the first one found will be set as the default ')
@click.pass_context
def gc(ctx, default_sender, default_ip_pool, default_unsubscribe_group, default_dynamic_template):
    """ get and format configuration information so its suitable for consumption by the patten_email class"""
    senders = ctx.invoke(gs, dump_std=False)
    ip_pools = ctx.invoke(gi, dump_std=False)
    asm = ctx.invoke(ga, dump_std=False)
    templates = ctx.invoke(gt, dump_std=False)

    auto_generated_config_dict = {}
    auto_generated_config_dict['api_key'] = ctx.obj.get('api_key')


    sender_config = {}
    for sender in senders:
        del sender['updated_at']
        del sender['created_at']
        del sender['locked']
        del sender['id']
        del sender['verified']
        del sender['country']
        sender_config[sender['nickname']] = sender
        sender_config[sender['nickname']]['from_address'] = sender.pop('from')

    sender_keys = list(sender_config.keys())
    if default_sender in sender_keys:
        sender_config['DEFAULT'] = sender_config[default_sender]
    else:
        if len(sender_config) > 0:
            sender_config['DEFAULT'] = sender_config[sender_keys[0]]

    auto_generated_config_dict['senders'] = sender_config


    ip_pool_config = {}
    for ip_pool in ip_pools:
        ip_pool_config[ip_pool['name']] = ip_pool

    ip_pool_keys = list(ip_pool_config.keys())
    if default_ip_pool in ip_pool_keys:
        ip_pool_config['DEFAULT'] = ip_pool_config[default_ip_pool]
    else:
        if len(ip_pool_config) > 0:
            ip_pool_config['DEFAULT'] = ip_pool_config[ip_pool_keys[0]]

    auto_generated_config_dict['ip_pools'] = ip_pool_config


    unsubscribe_groups_config = {}
    for unsubscribe_group in asm:
        unsubscribe_groups_config[unsubscribe_group['name']] = {}
        unsubscribe_groups_config[unsubscribe_group['name']]['id'] = unsubscribe_group['id']

    unsubscribe_group_keys = list(unsubscribe_groups_config.keys())
    if default_unsubscribe_group in unsubscribe_group_keys:
        unsubscribe_groups_config['DEFAULT'] = unsubscribe_groups_config[default_unsubscribe_group]
    else:
        if len(unsubscribe_groups_config) > 0:
            unsubscribe_groups_config['DEFAULT'] = unsubscribe_groups_config[unsubscribe_group_keys[0]]

    auto_generated_config_dict['unsubscribe_groups'] = unsubscribe_groups_config

    templates_config = {}
    for template in templates:
        templates_config[template['name']]= {}
        templates_config[template['name']]['id'] = template['id']
        templates_config[template['name']]['name'] = template['name']
        isolated_template_variables = ctx.invoke(gtv, template_id = template['id'], dump_std=False)
        templates_config[template['name']]['variables'] = isolated_template_variables


    template_keys = list(templates_config.keys())
    if default_dynamic_template in template_keys:
        templates_config['DEFAULT'] = templates_config[default_dynamic_template]
    else:
        if len(templates_config) > 0:
            templates_config['DEFAULT'] = templates_config[template_keys[0]]

    auto_generated_config_dict['email_templates'] = templates_config

    click.echo(json.dumps(auto_generated_config_dict))




@click.command()
@click.option('--dump-std', default=True )
@click.pass_context
def gs(ctx, dump_std):
    """ get sendgird senders """
    response = ctx.obj['sg_client'].senders.get()
    body = response.body.decode('utf-8')
    if dump_std:
        click.echo(body)
    return json.loads(body)


@click.command()
@click.option('--dump-std', default=True )
@click.pass_context
def ga(ctx, dump_std):
    """ get sendgird asms (unsubscribe groups)"""
    params = {}
    response = ctx.obj['sg_client'].asm.groups.get(query_params=params)
    body = response.body.decode('utf-8')
    if dump_std:
        click.echo(body)
    return json.loads(body)

@click.command()
@click.option('--dump-std', default=True )
@click.pass_context
def gt(ctx, dump_std):
    """ get sendgird dynamic templates """
    params = {'generations': 'dynamic'}
    response = ctx.obj['sg_client'].templates.get(query_params=params)
    body = response.body.decode('utf-8')
    if dump_std:
        click.echo(body)
    return json.loads(body)['templates']
    # return json.loads(response.body.decode('utf-8'))['templates']


@click.command()
@click.option('--dump-std', default=True )
@click.pass_context
def gi(ctx, dump_std):
    """ get sendgird ip pools """
    response = ctx.obj['sg_client'].ips.pools.get()
    body = response.body.decode('utf-8')
    if dump_std:
        click.echo(body)
    return json.loads(body)

@click.command()
@click.argument('template_id')
@click.option('--dump-std', default=True )
@click.pass_context
def gtd(ctx, template_id, dump_std):
    """ get details for a specific template """
    response = ctx.obj['sg_client'].templates._(template_id).get()
    body = json.loads(response.body.decode('utf-8'))
    # get the active version of the template
    template = None
    for version in body['versions']:
        if version['active'] == 1:
            template = version
            break
    if not template:
        # @todo convert logger
        pass
    del(body['versions'])
    body['template'] = template

    if dump_std:
        click.echo(body)
    return body

@click.command()
@click.argument('template_id')
@click.option('--dump-std', default=True )
@click.pass_context
def gtv(ctx, template_id, dump_std):
    """ get the variables defined in a specific template. """

    body = ctx.invoke(gtd, template_id = template_id, dump_std=False)
    if not body['template']:
        return []
    # Regular expression to find all Mustache variables
    # @todo the parsing could be better also variables in the subject line are not detected.
    content_variables = re.findall(r'{{\s*([^}]+)\s*}}', body['template']['plain_content'])
    subject_variables = []
    if 'subject' in body['template'].keys():
        subject_variables = re.findall(r'{{\s*([^}]+)\s*}}', body['template']['subject'])

    try:
        # these are defined with the asm config, @todo find a better mustache pattern the extra '{' is weird
        content_variables.remove('{unsubscribe')
        content_variables.remove('{unsubscribe_preferences')
    except:
        pass

    if dump_std:
        click.echo(f"content variables {content_variables}")
        click.echo(f"subject variables {subject_variables}")
    return content_variables + subject_variables


import os
import click
from sendgrid import SendGridAPIClient
from cli.get_config import gc, gs, ga, gt, gtd, gi, gtv


@click.group()
@click.pass_context
def pe_cli(ctx):
    """CLI interface into the sendgrid backend to run export SENDGRID_API_KEY environment variable"""
    api_key = os.getenv('SENDGRID_API_KEY', None)
    if not api_key:
        # @todo the key should be able to be supplied as an optional argument
        click.echo("missing SENDGRID_API_KEY environment variable")
        exit(1)
    sg = SendGridAPIClient(api_key=api_key)
    ctx.obj = {'sg_client': sg.client, 'api_key': api_key}


pe_cli.add_command(gc)
pe_cli.add_command(gs)
pe_cli.add_command(ga)
pe_cli.add_command(gt)
pe_cli.add_command(gi)
pe_cli.add_command(gtd)
pe_cli.add_command(gtv)

if __name__ == '__main__':
    pe_cli()



class MailSendFailure(Exception):
    def __init__(self, *args):
        if args:
            self.message = args[0]
        else:
            self.message = None

    def __str__(self):
        if self.message:
            return "MailSendFailure, {0} ".format(self.message)
        else:
            return "MailSendFailure attempts to send email failed"


class MissingAPIKey(Exception):
    def __init__(self, *args):
        if args:
            self.message = args[0]
        else:
            self.message = None

    def __str__(self):
        if self.message:
            return "MissingAPIKey, {0} ".format(self.message)
        else:
            return "MissingAPIKey a sendgrid API key must be provided to create an instance this class"


class MalformedConfiguration(Exception):
    def __init__(self, *args):
        if args:
            self.message = args[0]
        else:
            self.message = None

    def __str__(self):
        if self.message:
            return "MalformedConfiguration, {0} ".format(self.message)
        else:
            return "MalformedConfiguration supplied configuration is not valid or missing"
        

class BadSenderObject(Exception):
    def __init__(self, *args):
        if args:
            self.message = args[0]
        else:
            self.message = None

    def __str__(self):
        if self.message:
            return "BadSenderObject, {0} ".format(self.message)
        else:
            return 'BadSenderObject: sender parameter object must include "email" key' 